/* * This file is part of Harmattan Organiser GUI *
 *
 * Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
 * All rights reserved.
 *
 * Contact: Andrey Moiseenko <andrey.moiseenko@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef DEFAULTINVITATIONPLUGIN_H
#define DEFAULTINVITATIONPLUGIN_H

#include <invitationhandlerif.h>
#include "servicehandlerif.h"
#include <QtCore/QObject>

using namespace KCalCore;

/**
 * \brief DefaultInvitationPlugin class.
 *
 *  When no other plugin has been found for a particular notebook, then this plugin
 *  will try to send the invitation using QMF's "default account".
 *  Therefore this plugin, unlike others, will not have to use the accountId.
 */
class DefaultInvitationPlugin : public QObject, public InvitationHandlerInterface, public ServiceInterface
{
    Q_OBJECT
    Q_INTERFACES(InvitationHandlerInterface)
    Q_INTERFACES(ServiceInterface)

public:
    //! \brief DefaultInvitationPlugin constructor class.
    DefaultInvitationPlugin();

    //! \brief DefaultInvitationPlugin destructor class.
    ~DefaultInvitationPlugin();

    //! \reimp InvitationHandler KCalCore
    bool sendInvitation(const QString &accountId, const QString &notebookId, const Incidence::Ptr &invitation, const QString &body);
    bool sendUpdate(const QString &accountId, const Incidence::Ptr &invitation, const QString &body);
    bool sendResponse(const QString &accountId, const Incidence::Ptr &invitation, const QString &body);
    QString pluginName() const;
    //! \reimp_end

    //! \reimp ServiceHandler mKCal
    QString icon() const;

    QString uiName() const;

    bool multiCalendar() const;

    QString emailAddress(const mKCal::Notebook::Ptr &notebook);

    QString displayName(const mKCal::Notebook::Ptr &notebook) const;

    bool downloadAttachment(const mKCal::Notebook::Ptr &notebook, const QString &uri, const QString &path);

    bool deleteAttachment(const mKCal::Notebook::Ptr &notebook, const Incidence::Ptr &incidence, const QString &uri);

    bool shareNotebook(const mKCal::Notebook::Ptr &notebook, const QStringList &sharedWith);

    QStringList sharedWith(const mKCal::Notebook::Ptr &notebook);

    QString serviceName() const;

    QString defaultNotebook() const;

    bool checkProductId(const QString &prodId) const;

    ErrorCode error() const;

    //! \reimp_end

private:
  //@cond PRIVATE
  Q_DISABLE_COPY( DefaultInvitationPlugin )
  class Private;
  Private *const d;
  //@endcond
};

#endif
