/* * This file is part of Harmattan Organiser GUI *
 *
 * Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
 * All rights reserved.
 *
 * Contact: Andrey Moiseenko <andrey.moiseenko@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */


#include "transmitemail.h"

TransmitHelper::TransmitHelper(QMailAccountId id) : mId(id)
{
    mAction = new QMailTransmitAction(this);
    Q_ASSERT(mAction);

    connect(mAction, SIGNAL(activityChanged(QMailServiceAction::Activity)), this, SLOT(changeActivity(QMailServiceAction::Activity)));
}

void TransmitHelper::transmit()
{
    qDebug() << "TransmitHelper::transmit()" << mId;
	mAction->transmitMessages(mId);
}

void TransmitHelper::changeActivity(QMailServiceAction::Activity a)
{
    switch (a)
    {
        case QMailServiceAction::Pending:
            qDebug() << "Pending request to server...";
            break;

        case QMailServiceAction::InProgress:
            qDebug() << "Request to server in progress...";
            break;

        case QMailServiceAction::Successful:
            qDebug() << "Request to server has been completed successfull!";
            emit done();
            break;

        case QMailServiceAction::Failed:
            qDebug() << "Request to server failed!";
            emit done();
            break;

        default:
            Q_ASSERT(false);
            break;
    }
}
